from __future__ import annotations

import asyncio
import dataclasses
import logging
import typing as t

import yaml

from airpixel import framework, monitoring, logging_config


log = logging.getLogger("airpixel")


@dataclasses.dataclass
class Config:
    framework_config: framework.Config
    monitoring_config: monitoring.Config

    @classmethod
    def from_dict(cls, dict_: t.Dict[str, t.Any]) -> Config:
        return cls(
            framework.Config.from_dict(dict_["framework"]),
            monitoring.Config.from_dict(dict_["monitoring"]),
        )

    @classmethod
    def load(cls, file_name: str) -> Config:
        with open(file_name) as file_:
            return cls.from_dict(yaml.safe_load(file_))


async def main() -> None:
    logging_config.load("airpixel.yaml")
    config = Config.load("airpixel.yaml")
    framework_app = framework.Application(config.framework_config)
    monitoring_app = monitoring.Application(config.monitoring_config)
    try:
        await asyncio.gather(framework_app.run_forever(), monitoring_app.run_forever())
    except KeyboardInterrupt:
        log.info("Application shut down by user (keyboard interrupt)")


asyncio.run(main())



import logging.config

import yaml


def load(filename: str) -> None:
    with open(filename) as file_:
        config = yaml.safe_load(file_).get("logging")
    if config is None:
        return
    logging.config.dictConfig(config)



import os
import numpy as np  # type: ignore
import io
import time

from airpixel import client

mon_client = client.MonitorClient(
    os.path.join(os.path.dirname(os.path.dirname(__file__)), "monitoring_uds")
)

while True:
    arr = np.random.random(1000)
    file_ = io.BytesIO()
    np.save(file_, arr, False)
    mon_client.send_data("random", file_.getvalue())
    time.sleep(0.01)



import sys

import time
import socket


def main() -> None:
    name = sys.argv[1]

    with socket.create_connection(("0.0.0.0", 50000)) as sock:
        sock.send(int.to_bytes(54325, 2, "big") + bytes(name, "utf-8") + b"\n")
        port_bytes = sock.recv(8)
        port = int.from_bytes(port_bytes, byteorder="big")

    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.settimeout(0)
    while True:
        time.sleep(2)
        sock.sendto(b"1 1", ("0.0.0.0", port))


if __name__ == "__main__":
    main()



from __future__ import annotations

import asyncio
import dataclasses
import enum
import logging
import logging.config
import socket
import time
import typing as t

import yaml

from airpixel import logging_config


log = logging.getLogger(__name__)


class MonitoringError(Exception):
    pass


class PackageParsingError(MonitoringError):
    pass


class PackageSerializationError(MonitoringError):
    pass


class CommandParseError(MonitoringError):
    pass


class CommandError(MonitoringError):
    pass


class CommandResponseParseError(MonitoringError):
    pass


@enum.unique
class CommandVerb(str, enum.Enum):
    SUBSCRIBE = "sub"
    UNSUBSCRIBE = "unsub"
    CONNECT = "conn"


@enum.unique
class CommandResponseType(bytes, enum.Enum):
    ERROR = b"err"
    SUCCESS = b"acc"


@dataclasses.dataclass
class CommandResponse:
    response: CommandResponseType
    info: str
    SEPERATOR = b":"

    def to_bytes(self) -> bytes:
        return self.response.value + self.SEPERATOR + bytes(self.info, "utf-8")

    @classmethod
    def from_bytes(cls, data: bytes) -> CommandResponse:
        try:
            response, info = data.split(cls.SEPERATOR)
        except ValueError as e:
            raise CommandResponseParseError("Invalid command response") from e
        try:
            response = CommandResponseType(response)
        except ValueError as e:
            raise CommandResponseParseError(
                f"Invalid command response: {str(response, 'utf-8')}"
            ) from e
        return cls(response, str(info, "utf-8"))


@dataclasses.dataclass
class Command:
    verb: CommandVerb
    arg: str

    @classmethod
    def from_bytes(cls, data: bytes) -> Command:
        try:
            verb_str, arg = str(data, "utf-8").split(" ")
        except ValueError as e:
            raise CommandParseError("Invalid command") from e
        try:
            verb = CommandVerb(verb_str)
        except ValueError as e:
            raise CommandParseError("Invalid command") from e
        return cls(verb, arg.strip())

    def to_bytes(self) -> bytes:
        return bytes(self.verb.value, "utf-8") + b" " + bytes(self.arg, "utf-8") + b"\n"


@dataclasses.dataclass
class Package:
    stream_id: str
    data: bytes

    SEPARATOR = b"\x00"

    @classmethod
    def from_bytes(cls, raw_data: bytes) -> Package:
        try:
            header, data = raw_data.split(cls.SEPARATOR, 1)
        except ValueError as e:
            raise PackageParsingError("Invalid package") from e
        if not header:
            raise PackageParsingError("The package is invalid: No stream ID")
        return cls(str(header, "utf-8"), data)

    def to_bytes(self) -> bytes:
        if not self.stream_id:
            raise PackageSerializationError("stream_id can't be empty")
        stream_id_bytes = bytes(self.stream_id, "utf-8")
        return stream_id_bytes + self.SEPARATOR + self.data


class DispachProtocol(asyncio.DatagramProtocol):
    def __init__(self, monitoring_server: Server):
        super().__init__()
        self._monitoring_server = monitoring_server

    def datagram_received(self, data: bytes, addr: t.Tuple[str, int]) -> None:
        try:
            package = Package.from_bytes(data)
        except PackageParsingError:
            log.debug("Invalid package", exc_info=True)
            return
        self._monitoring_server.dispatch_to_monitors(package.stream_id, data)


class KeepaliveProtocol(asyncio.DatagramProtocol):
    def __init__(self, monitoring_server: Server):
        super().__init__()
        self._monitoring_server = monitoring_server

    def datagram_received(self, data: bytes, addr: t.Tuple[str, int]) -> None:
        ip_address, _ = addr
        self._monitoring_server.message_from(ip_address)


@dataclasses.dataclass
class Device:
    ip_address: str
    udp_port: int
    last_message: float = 0
    subscriptions: t.Dict[str, Stream] = dataclasses.field(default_factory=dict)

    def subscribe_to(self, stream: Stream) -> None:
        self.subscriptions[stream.stream_id] = stream
        stream.add_subscriber(self)

    def unsubscribe_from(self, stream: Stream) -> None:
        try:
            subscription = self.subscriptions.pop(stream.stream_id)
        except KeyError:
            return
        subscription.remove_subscriber(self)

    def unsubscribe_all(self) -> t.List[Stream]:
        for subscription in self.subscriptions.values():
            subscription.remove_subscriber(self)
        old_subscriptions = [
            subscription for subscription in self.subscriptions.values()
        ]
        self.subscriptions = {}
        return old_subscriptions

    def address(self) -> t.Tuple[str, int]:
        return (self.ip_address, self.udp_port)

    def heartbeat(self) -> None:
        self.last_message = time.time()


@dataclasses.dataclass
class Stream:
    stream_id: str
    subscribers: t.Dict[str, Device] = dataclasses.field(default_factory=dict)

    def add_subscriber(self, monitor: Device) -> None:
        self.subscribers[monitor.ip_address] = monitor

    def remove_subscriber(self, monitor: Device) -> None:
        del self.subscribers[monitor.ip_address]

    def has_subscribers(self) -> bool:
        return bool(self.subscribers)


class Server:
    def __init__(self, subscription_timeout: int = 3):
        self.subscription_timeout = subscription_timeout
        self._devices: t.Dict[str, Device] = {}
        self._streams: t.Dict[str, Stream] = {}
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket.settimeout(0)

    def connect(self, ip_address: str, port: int) -> None:
        device = self._devices.setdefault(ip_address, Device(ip_address, port))
        device.heartbeat()
        log.debug("Monitor %s connected", ip_address)

    def dispatch_to_monitors(self, stream_id: str, data: bytes) -> None:
        log.debug("dispatching for stream %s", stream_id)
        try:
            stream = self._streams[stream_id]
        except KeyError:
            log.debug("No subscriptions for %s", stream_id)
            return
        for subscriber in stream.subscribers.values():
            try:
                self.socket.sendto(data, subscriber.address())
            except OSError:
                pass
            log.debug("Sent data for '%s' to %s", stream_id, subscriber.ip_address)

    def subscribe_to_stream(self, ip_address: str, stream_id: str) -> None:
        try:
            device = self._devices[ip_address]
        except KeyError:
            return
        stream = self._streams.setdefault(stream_id, Stream(stream_id))
        device.subscribe_to(stream)
        log.debug("%s subscribed to stream %s", ip_address, stream_id)

    def unsubscribe_from_stream(self, ip_address: str, stream_id: str) -> None:
        try:
            device = self._devices[ip_address]
            stream = self._streams[stream_id]
        except KeyError:
            return
        del self._devices[ip_address]
        device.unsubscribe_from(stream)
        self._clean_stream(stream)

    def _clean_stream(self, stream: Stream) -> None:
        if not stream.has_subscribers():
            del self._streams[stream.stream_id]

    def message_from(self, ip_address: str) -> None:
        try:
            device = self._devices[ip_address]
        except KeyError:
            return
        device.heartbeat()

    def purge_subscriptions(self) -> None:
        now = time.time()
        to_kill: t.Set[str] = set()
        for device in self._devices.values():
            if now - device.last_message < self.subscription_timeout:
                continue
            to_kill.add(device.ip_address)
            streams = device.unsubscribe_all()
            for stream in streams:
                self._clean_stream(stream)
        for ip_address in to_kill:
            log.debug("Killed Monitor %s", ip_address)
            del self._devices[ip_address]

    async def purge_forever(self) -> None:
        while True:
            self.purge_subscriptions()
            await asyncio.sleep(self.subscription_timeout / 4)


class ConnectionProtocol(asyncio.Protocol):
    PORT_SIZE = 2
    SEPPERATOR = b"\n"
    DEFAULT_RESPONSE = "acc"
    transport: asyncio.Transport

    def __init__(self, monitoring_server: Server, keepalive_port: int):
        super().__init__()
        self._monitoring_server = monitoring_server
        self._current_package = b""
        self._keepalive_port = keepalive_port

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        self.transport = t.cast(asyncio.Transport, transport)

    def _subscribe(self, arg: str) -> str:
        stream_id = arg
        ip_address, _ = self.transport.get_extra_info("peername")
        self._monitoring_server.subscribe_to_stream(ip_address, stream_id)
        return self.DEFAULT_RESPONSE

    def _unsubscribe(self, arg: str) -> str:
        stream_id = arg
        ip_address, _ = self.transport.get_extra_info("peername")
        self._monitoring_server.unsubscribe_from_stream(ip_address, stream_id)
        return self.DEFAULT_RESPONSE

    def _connect(self, args: str) -> str:
        try:
            udp_port = int(args)
        except ValueError:
            raise CommandError("port needs to be an int")
        ip_address, _ = self.transport.get_extra_info("peername")
        self._monitoring_server.connect(ip_address, udp_port)
        return str(self._keepalive_port)

    def execute_command(self, command: Command) -> str:
        if command.verb == CommandVerb.SUBSCRIBE:
            return self._subscribe(command.arg)
        if command.verb == CommandVerb.UNSUBSCRIBE:
            return self._unsubscribe(command.arg)
        if command.verb == CommandVerb.CONNECT:
            return self._connect(command.arg)
        raise CommandError("unrecognized command verb")

    def respond_error(self, error: Exception) -> None:
        self.transport.write(
            CommandResponse(CommandResponseType.ERROR, str(error)).to_bytes()
        )
        self.transport.close()

    def respond_success(self, data: str) -> None:
        self.transport.write(
            CommandResponse(CommandResponseType.SUCCESS, data).to_bytes()
        )
        self.transport.close()

    def data_received(self, data: bytes) -> None:
        *packages, self._current_package = (self._current_package + data).split(
            self.SEPPERATOR
        )
        if not packages:
            return
        try:
            command = Command.from_bytes(data)
        except CommandParseError as e:
            self.respond_error(e)
            return
        try:
            response = self.execute_command(command)
        except CommandError as e:
            self.respond_error(e)
        else:
            self.respond_success(response)


@dataclasses.dataclass
class Config:
    address: str
    port: int
    unix_socket: str

    @classmethod
    def from_dict(cls, dict_: t.Dict[str, t.Any]) -> Config:
        return cls(dict_["address"], dict_["port"], dict_["unix_socket"],)

    @classmethod
    def load(cls, file_name: str) -> Config:
        with open(file_name) as file_:
            return cls.from_dict(yaml.safe_load(file_)["monitoring"])


class Application:
    def __init__(self, config: Config):
        self.config = config
        self.monitoring_server = Server()

    async def run_forever(self) -> None:
        loop = asyncio.get_running_loop()

        transport, _ = await loop.create_datagram_endpoint(
            lambda: KeepaliveProtocol(self.monitoring_server),
            local_addr=(self.config.address, 0),
            family=socket.AF_INET,
        )
        _, keepalive_port = transport.get_extra_info("sockname")

        log.info(
            "Monitoring keepalive endpoint up on %(keepalive_port)s",
            {"keepalive_port": keepalive_port},
        )

        await loop.create_datagram_endpoint(
            lambda: DispachProtocol(self.monitoring_server),
            local_addr=t.cast(t.Any, self.config.unix_socket),
            family=socket.AF_UNIX,
        )

        log.info(
            "Monitoring dispatch endpoint up on %(unix_socket)s",
            {"unix_socket": self.config.unix_socket},
        )

        server = await loop.create_server(
            lambda: ConnectionProtocol(self.monitoring_server, keepalive_port),
            self.config.address,
            self.config.port,
        )

        log.info(
            "Monitoring server created on %(address)s:%(port)s",
            {"address": self.config.address, "port": self.config.port},
        )

        async with server:
            await asyncio.gather(
                server.serve_forever(), self.monitoring_server.purge_forever()
            )


def main() -> None:
    logging_config.load("airpixel.yaml")
    config = Config.load("airpixel.yaml")

    log.info("Monitoring configuration loaded")

    app = Application(config)
    try:
        asyncio.run(app.run_forever())
    except KeyboardInterrupt:
        log.info("Application shut down by user (keyboard interrupt)")


if __name__ == "__main__":
    main()



from __future__ import annotations

import atexit
import asyncio
import dataclasses
import logging
import socket
import subprocess
import time
import typing as t

import yaml


logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(__name__)

BYTEORDER = "big"


class FrameworkException(Exception):
    pass


class KeepaliveProtocol(asyncio.DatagramProtocol):
    def __init__(self, process_registration: ProcessRegistration):
        super().__init__()
        self._process_registration = process_registration

    def datagram_received(self, data: bytes, addr: t.Tuple[str, int]) -> None:
        ip_address, _ = addr
        frames, rendered = (int(n) for n in str(data, "utf-8").split())
        if frames:
            log.info(
                "%s: Received: %s Shown: %s Ratio: %s",
                ip_address,
                frames,
                rendered,
                rendered / frames,
            )
        self._process_registration.response_from(ip_address)


def _subprocess_factory(command: str) -> subprocess.Popen:
    return subprocess.Popen("exec " + command, text=True, shell=True)


@dataclasses.dataclass
class ProcessMeta:
    process: subprocess.Popen
    ip_address: str
    device_id: str
    last_response: float


class ProcessRegistration:
    def __init__(
        self,
        device_configs: t.Iterable[DeviceConfig],
        subprocess_factory: t.Callable[[str], subprocess.Popen] = _subprocess_factory,
        timeout: float = 3,
    ):
        self._commands = {
            device_config.device_id: device_config.command_template
            for device_config in device_configs
        }
        self._subprocess_factory = subprocess_factory
        self._timeout = timeout
        self._processes: t.Dict[str, ProcessMeta] = {}
        atexit.register(self.cleanup)

    def _kill_process(self, ip_address: str) -> None:
        if ip_address not in self._processes:
            return
        self._processes[ip_address].process.kill()
        self._processes[ip_address].process.communicate()
        del self._processes[ip_address]

    def response_from(self, ip_address: str) -> None:
        try:
            self._processes[ip_address].last_response = time.time()
        except KeyError:
            pass

    def purge_processes(self) -> None:
        now = time.time()
        dead_processes = {
            ip
            for ip, process in self._processes.items()
            if now - process.last_response >= self._timeout
        }
        for ip_address in dead_processes:
            log.info("Killing process for %s.", ip_address)
            self._kill_process(ip_address)

    async def purge_forever(self) -> None:
        while True:
            self.purge_processes()
            await asyncio.sleep(self._timeout / 4)

    def launch_for(self, device_id: str, ip_address: str, streaming_port: int) -> None:
        try:
            base_command = self._commands[device_id]
        except KeyError:
            log.warning("No process configured for device ID %s", device_id)
            return
        try:
            base_command = base_command.format(
                ip_address=ip_address, port=str(streaming_port)
            )
        except KeyError:
            log.warning(
                "Invalid format string for subprocess command for device %s", device_id
            )
            return
        log.info("Launching process for device %s: `%s`", device_id, base_command)
        self._kill_process(ip_address)
        self._processes[ip_address] = ProcessMeta(
            self._subprocess_factory(base_command), ip_address, device_id, time.time()
        )

    def cleanup(self) -> None:
        for process_meta in self._processes.values():
            process_meta.process.kill()


class ConnectionProtocol(asyncio.Protocol):
    PORT_SIZE = 2
    SEPPERATOR = b"\n"
    transport: asyncio.Transport

    def __init__(self, process_registration: ProcessRegistration, response_port: int):
        super().__init__()
        self._process_registration = process_registration
        self._current_package = b""
        self.response_port = response_port

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        self.transport = t.cast(asyncio.Transport, transport)

    def _register_device(self, registration_bytes: bytes) -> None:
        port = int.from_bytes(registration_bytes[: self.PORT_SIZE], BYTEORDER)
        device_id = str(registration_bytes[self.PORT_SIZE :], "utf-8")
        ip_address, _ = self.transport.get_extra_info("peername")
        self._process_registration.launch_for(device_id, ip_address, port)
        self.transport.write(
            int.to_bytes(self.response_port, self.PORT_SIZE, BYTEORDER)
        )
        log.info("Registered device %s", device_id)

    def data_received(self, data: bytes) -> None:
        *packages, self._current_package = (self._current_package + data).split(
            self.SEPPERATOR
        )
        if not packages:
            return
        log.info("Received %r", packages)
        self._register_device(packages[0])
        self.transport.close()


@dataclasses.dataclass
class Config:
    address: str
    port: int
    udp_port: int
    devices: t.List[DeviceConfig]

    @classmethod
    def from_dict(cls, dict_: t.Dict[str, t.Any]) -> Config:
        return cls(
            dict_["address"],
            dict_["port"],
            dict_["udp_port"],
            [DeviceConfig.from_dict(d) for d in dict_["devices"]],
        )

    @classmethod
    def load(cls, file_name: str) -> Config:
        with open(file_name) as file_:
            return cls.from_dict(yaml.safe_load(file_)["framework"])


@dataclasses.dataclass
class DeviceConfig:
    device_id: str
    command_template: str

    @classmethod
    def from_dict(cls, dict_: t.Dict[str, t.Any]) -> DeviceConfig:
        return cls(dict_["device_id"], dict_["command_template"],)


class Application:
    def __init__(self, config: Config):
        self.config = config
        self.process_registration = ProcessRegistration(config.devices)

    async def run_forever(self) -> None:
        loop = asyncio.get_running_loop()

        await loop.create_datagram_endpoint(
            lambda: KeepaliveProtocol(self.process_registration),
            local_addr=(self.config.address, self.config.udp_port),
            family=socket.AF_INET,
        )

        server = await loop.create_server(
            lambda: ConnectionProtocol(self.process_registration, self.config.udp_port),
            self.config.address,
            self.config.port,
        )

        async with server:
            await asyncio.gather(
                server.serve_forever(), self.process_registration.purge_forever()
            )


def main() -> None:
    config = Config.load("airpixel.yaml")

    log.info(config)

    app = Application(config)
    try:
        asyncio.run(app.run_forever())
    except KeyboardInterrupt:
        log.info("Application shut down by user (keyboard interrupt)")


if __name__ == "__main__":
    main()



from __future__ import annotations

import asyncio
import socket
import typing as t
import dataclasses
import io
import threading

import yaml

import pyqtgraph as pg  # type: ignore
import numpy as np  # type: ignore
from pyqtgraph.Qt import QtGui  # type: ignore
from PyQt5 import QtCore  # type: ignore

from airpixel import monitoring


_PlotDict = t.Dict[str, "SimplePlot"]


class PlotProtocol(asyncio.DatagramProtocol):
    def __init__(self, plots: _PlotDict):
        self.plots = plots

    def datagram_received(self, data: bytes, addr: t.Tuple[str, int]) -> None:
        package = monitoring.Package.from_bytes(data)

        if package.stream_id not in self.plots:
            return

        memfile = io.BytesIO()
        memfile.write(package.data)
        memfile.seek(0)
        array = np.load(memfile)

        self.plots[package.stream_id].new_data.emit(array)


class MonitorServer:
    def __init__(self, ip_address: str, port: int, plots: _PlotDict):
        self.ip_address = ip_address
        self.port = port
        self.plots = plots
        self.local_port = 0
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket.settimeout(0)

    async def run_forever(self) -> None:
        loop = asyncio.get_running_loop()
        transport, _ = await loop.create_datagram_endpoint(
            lambda: PlotProtocol(self.plots),
            local_addr=("0.0.0.0", 0),
            family=socket.AF_INET,
        )
        _, self.local_port = transport.get_extra_info("sockname")

        keepalive_port = self.connect()

        for stream_id in self.plots:
            self.subscribe(stream_id)

        while True:
            self.socket.sendto(b"", (self.ip_address, keepalive_port))
            await asyncio.sleep(1)

    def subscribe(self, stream_id: str) -> None:
        with socket.create_connection((self.ip_address, int(self.port))) as sock:
            sock.send(
                monitoring.Command(
                    monitoring.CommandVerb.SUBSCRIBE, stream_id
                ).to_bytes()
            )

    def connect(self) -> int:
        with socket.create_connection((self.ip_address, int(self.port))) as sock:
            sock.send(
                monitoring.Command(
                    monitoring.CommandVerb.CONNECT, str(self.local_port),
                ).to_bytes()
            )
            response_data = sock.recv(128)
        response = monitoring.CommandResponse.from_bytes(response_data)
        return int(response.info)

    def run(self) -> None:
        asyncio.run(self.run_forever())


@dataclasses.dataclass
class StreamConfig:
    name: str

    @classmethod
    def from_dict(cls, dict_: t.Dict[str, t.Any]) -> StreamConfig:
        return cls(dict_["name"])


@dataclasses.dataclass
class Config:
    server: str
    port: int
    streams: t.List[StreamConfig]

    @classmethod
    def from_dict(cls, dict_: t.Dict[str, t.Any]) -> Config:
        return cls(
            dict_["server"],
            dict_["port"],
            [StreamConfig.from_dict(dict_) for dict_ in dict_["streams"]],
        )

    @classmethod
    def load(cls, file_name: str) -> Config:
        with open(file_name) as file_:
            return cls.from_dict(yaml.safe_load(file_)["monitor"])


class SimplePlot(pg.PlotWidget):
    new_data = QtCore.pyqtSignal(np.ndarray)

    def __init__(self, stream_id: str):
        super().__init__(title=stream_id)
        self.stream_id = stream_id
        self._current_max_y = 0
        self._curve = self.plot(
            np.array([0]),
            np.array([]),
            pen="y",
            fillLevel=0,
            fillBrush="y",
            stepMode=True,
        )

    def _fit_plot(self, points: np.ndarray) -> None:
        self._current_max_y = max(max(points), self._current_max_y)
        self.setRange(yRange=(0, self._current_max_y))

    def plot_array(self, data: np.ndarray) -> None:
        self._curve.setData(np.arange(len(data) + 1), data)
        self._fit_plot(data)


class Application:
    def __init__(self, config: Config, q_app: QtGui.QApplication):
        self.config = config
        self.plots: _PlotDict = {}
        self.server = MonitorServer(self.config.server, self.config.port, self.plots)

        self.q_app = q_app
        self.top_widget = QtGui.QWidget()
        self.top_widget.resize(1800, 900)

        self.layout = QtGui.QGridLayout()
        self.top_widget.setLayout(self.layout)
        for i, stream in enumerate(self.config.streams):
            self.plots[stream.name] = SimplePlot(stream.name)
            self.plots[stream.name].new_data.connect(self.plots[stream.name].plot_array)
            self.layout.addWidget(self.plots[stream.name], i, 0)

    def run_forever(self) -> None:
        server_thread = threading.Thread(target=self.server.run, daemon=True)
        server_thread.start()

        self.top_widget.show()

        self.q_app.exec_()


def main() -> None:
    config = Config.load("monitor.yaml")

    q_app = QtGui.QApplication([])
    app = Application(config, q_app)
    try:
        app.run_forever()
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['airpixel']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.19.1,<2.0.0', 'pyyaml>=5.3.1,<6.0.0']

extras_require = \
{'lint': ['black>=19.10b0,<20.0', 'flake8>=3.8.1,<4.0.0'],
 'monitoring': ['pyqtgraph>=0.11.0,<0.12.0', 'pyqt5>=5.15.0,<6.0.0'],
 'test': ['pytest>=5.4.2,<6.0.0'],
 'typecheck': ['mypy>=0.770,<0.771']}

setup_kwargs = {
    'name': 'airpixel',
    'version': '0.9',
    'description': 'Controll LEDs with arduinoNanoIOT and Python',
    'long_description': None,
    'author': None,
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import abc
import io
import socket
import typing as t

import numpy as np  # type: ignore

from airpixel import gamma_table, monitoring


class UDPConstants:
    ENCODING_BYTEORDER = "big"
    FRAME_NUMBER_BYTES = 8
    BITS_IN_BYTE = 8
    MAX_PACKET_SIZE = 65507


class ColorMethod:
    @staticmethod
    @abc.abstractmethod
    def to_bytes(pixel: "Pixel") -> np.ndarray:
        pass


class ColorMethodRGB(ColorMethod):
    @staticmethod
    @abc.abstractmethod
    def to_bytes(pixel: "Pixel") -> np.ndarray:
        return pixel.values


class ColorMethodRGBW(ColorMethod):
    # For now just r, g, b, 0.
    # This might have a better solution:
    # http://www.mirlab.org/conference_papers/International_Conference/ICASSP%202014/papers/p1214-lee.pdf
    @staticmethod
    @abc.abstractmethod
    def to_bytes(pixel: "Pixel") -> np.ndarray:
        return np.append(pixel.values, [0])


class ColorMethodGRB(ColorMethod):
    @staticmethod
    @abc.abstractmethod
    def to_bytes(pixel: "Pixel") -> np.ndarray:
        red, green, blue = pixel.values
        return np.array((green, red, blue))


class ColorMethodGRBW(ColorMethod):
    @staticmethod
    @abc.abstractmethod
    def to_bytes(pixel: "Pixel") -> np.ndarray:
        red, green, blue, white = ColorMethodRGBW.to_bytes(pixel)
        return np.array((green, red, blue, white))


class Pixel:
    def __init__(self, red: float, green: float, blue: float) -> None:
        self.values = np.array((red, green, blue))

    def __repr__(self) -> str:
        return "P{:.2}|{:.2}|{:.2}".format(
            float(self.values[0]), float(self.values[1]), float(self.values[2])
        )


class AirClient:
    def __init__(
        self,
        remote_ip: str,
        remote_port: int,
        color_method: t.Type[ColorMethod] = ColorMethodGRB,
    ) -> None:
        self.remote_ip = remote_ip
        self.remote_port = remote_port
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket.settimeout(0)
        self.frame_number = 0
        self.color_method = color_method

    def send_bytes(self, message: bytes) -> None:
        try:
            self.socket.sendto(message, (self.remote_ip, self.remote_port))
        except OSError:
            pass

    def show_frame(self, frame: t.List[Pixel]) -> None:
        raw_pixels = np.concatenate(
            [self.color_method.to_bytes(pixel) for pixel in frame]
        )
        raw_pixels = gamma_table.GAMMA_TABLE[(raw_pixels * 255).astype("uint8")]
        frame_number = self.frame_number.to_bytes(
            UDPConstants.FRAME_NUMBER_BYTES, byteorder=UDPConstants.ENCODING_BYTEORDER
        )
        self.send_bytes(frame_number + bytes(raw_pixels))
        self.frame_number += 1


class MonitorClient:
    def __init__(self, socket_address: str):
        self.socket_address = socket_address
        self.socket = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
        self.socket.settimeout(0)

    def send_bytes(self, message: bytes) -> None:
        try:
            self.socket.sendto(message, self.socket_address)
        except OSError:
            pass

    def send_data(self, stream_id: str, data: bytes) -> None:
        bytes_ = monitoring.Package(stream_id, data).to_bytes()
        self.send_bytes(bytes_)

    def send_np_array(self, stream_id: str, data: np.array) -> None:
        file_ = io.BytesIO()
        np.save(file_, data, False)
        self.send_data(stream_id, file_.getvalue())
